#include "GameBuilders.h"

CommonBuilder::CommonBuilder(float width, float height, float room_size) :
        m_width(width), m_height(height), m_room_size(room_size){
}

void CommonBuilder::create_context(float dynamic_objects_ratio, int players) {
    std::vector<Room*> empty_rooms_buffer;
    for (auto& row : m_rooms)
        for (auto& room : row)
            if (room != nullptr)
                empty_rooms_buffer.emplace_back(room);
    //в рандомное место располагаем пакмана
    float id = std::rand() % empty_rooms_buffer.size();
    m_context.pacman.set_location(empty_rooms_buffer[id]);
    auto it = std::next(empty_rooms_buffer.begin(), id);
    empty_rooms_buffer.erase(it);
    if (players == 2) {
        float id2 = std::rand() % empty_rooms_buffer.size();
        m_context.pacman2.set_location(empty_rooms_buffer[id2]);
        auto it2 = std::next(empty_rooms_buffer.begin(), id2);
        empty_rooms_buffer.erase(it2);
    }

    //располагаем врагов
    size_t rooms_row = static_cast<size_t>(m_width / m_room_size);
    size_t rooms_col = static_cast<size_t>(m_height / m_room_size);
    for (size_t i = 0; i < rooms_row * rooms_col * dynamic_objects_ratio; i++) {
        id = std::rand() % empty_rooms_buffer.size();
        auto enemy = std::make_unique<Enemy>();
        enemy->set_location(empty_rooms_buffer.at(id));
        enemy->set_texture(&EnemyTexture::instance());
        m_context.dynamic_objects.emplace_back(std::move(enemy));
    }

    //располагаем еду
    for (auto it : empty_rooms_buffer) {
        auto food = std::make_unique<Food>();
        food->set_location(it);
        food->set_texture(&FoodTexture::instance());

        if (rand() % 100 < 10) food->is_screamer = true;
        m_context.static_objects.emplace_back(std::move(food));
    }
    empty_rooms_buffer.clear();
}

void CommonBuilder::create_state(IStateManager &state_manager, const sf::VideoMode &mode, const std::string &window_title) {
    m_game_state = std::make_unique<GameState>(state_manager, mode, window_title);
}

void CommonBuilder::set_all_to_state() {
    std::vector<Room*> rooms;
    for (auto& row : m_rooms)
        for (auto& col : row)
            rooms.emplace_back(col);
    Maze maze(rooms);
    m_game_state->set_maze(std::move(maze));
    m_game_state->set_context(std::move(m_context));
}

std::unique_ptr<GameState> CommonBuilder::get_game() {
    return std::move(m_game_state);
}

void SimpleBuilder::create_rooms() {
    size_t rows = m_height / m_room_size;
    size_t cols = m_width / m_room_size;
    int room_size = static_cast<int>(m_room_size);
    sf::Vector2f left_pos = sf::Vector2f{ (m_width - m_room_size * cols + m_room_size)/2, (m_height - m_room_size * rows + m_room_size)/2};
    for (int row = 0; row < rows * room_size; row += room_size) {
        std::vector<Room*> row_vec;
        for (size_t col = 0; col < cols * room_size; col += room_size) {
            Room* room = new Room(m_room_size);
            room->set_position({col + left_pos.x, row + left_pos.y});
            row_vec.emplace_back(room);
        }
        m_rooms.emplace_back(std::move(row_vec));
    }
}

void SimpleBuilder::set_rooms_sides() {
    for (size_t i = 0; i < m_rooms.size(); i++) {
        for (size_t j = 0; j < m_rooms.begin()->size(); j++) {
            if (i == 0) {
                auto up_wall = std::make_shared<Wall>(*m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::UP, std::move(up_wall));
            }
            else {
                auto pass = std::make_shared<Pass>(*m_rooms[i - 1][j], *m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::UP, pass);
                m_rooms[i - 1][j]->set_side(Room::DOWN, pass);
            }
            if (j == 0) {
                auto left_wall = std::make_shared<Wall>(*m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::LEFT, std::move(left_wall));
            }
            else {
                auto pass = std::make_shared<Pass>(*m_rooms[i][j - 1], *m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::LEFT, pass);
                m_rooms[i][j - 1]->set_side(Room::RIGHT, pass);
            }
            if (i == m_rooms.size() - 1) {
                auto down_wall = std::make_shared<Wall>(*m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::DOWN, std::move(down_wall));
            }
            if (j == m_rooms.begin()->size() - 1) {
                auto right_wall = std::make_shared<Wall>(*m_rooms[i][j]);
                m_rooms[i][j]->set_side(Room::RIGHT, std::move(right_wall));
            }
        }
    }

}

void ComplexBuilder::create_rooms() {
    size_t room_quantity_row = (m_height / m_room_size);
    size_t room_quantity_col = (m_width / m_room_size);

    int room_size = static_cast<int>(m_room_size);
    auto starting_point = sf::Vector2f{(m_width - m_room_size*room_quantity_col)/2 + m_room_size/2, (m_height - m_room_size*room_quantity_row)/2 + + m_room_size/2};

    for (int row = 0; row < room_quantity_row * room_size; row += room_size) {
        std::vector<Room*> vector_row;
        for (int col = 0; col < room_quantity_col * room_size; col += room_size) {
            auto room = new Room(m_room_size);
            room->set_position(sf::Vector2f {(col + starting_point.x), (row + starting_point.y)});
            vector_row.emplace_back(std::move(room));
        }
        m_rooms.emplace_back(std::move(vector_row));
    }
}

void ComplexBuilder::set_rooms_sides() {
    size_t cols = m_rooms[0].size();
    size_t rows = m_rooms.size();
    for (int j = 0; j < cols; j++) {
        m_rooms[0][j]->set_side(Room::Direction::UP, std::make_unique<Wall>(*m_rooms[0][j]));
    }
    for (int j = 0; j < cols; j++) {
        m_rooms[rows - 1][j]->set_side(Room::Direction::DOWN, std::make_unique<Wall>(*m_rooms[rows - 1][j]));
        if (j != cols - 1) {
            m_rooms[rows  - 1][j]->set_side(Room::Direction::RIGHT, std::make_unique<Pass>(*m_rooms[rows - 1][j], *m_rooms[rows - 1][j+1]));
            m_rooms[rows - 1][j+1]->set_side(Room::Direction::LEFT, std::make_unique<Pass>(*m_rooms[rows - 1][j+1], *m_rooms[rows - 1][j]));
        }
    }
    for (int i = 0; i < rows ; i++) {
        m_rooms[i][0]->set_side(Room::Direction::LEFT, std::make_unique<Wall>(*m_rooms[i][0]));
    }
    for (int i = 0; i < rows; i++) {
        m_rooms[i][cols - 1]->set_side(Room::Direction::RIGHT, std::make_unique<Wall>(*m_rooms[i][cols - 1]));
    }

    std::vector<int> row(cols);
    int count = 0;
    for (; count < cols; row[count] = count, count++);
    for (int i = 0; i < rows - 1; i++) {
        for (int j = 0; j < cols - 1; j++) {
            if (rand() % 2) {
                m_rooms[i][j]->set_side(Room::Direction::RIGHT, std::make_unique<Wall>(*m_rooms[i][j]));
                m_rooms[i][j+1]->set_side(Room::Direction::LEFT, std::make_unique<Wall>(*m_rooms[i][j+1]));
            }
            else {
                m_rooms[i][j]->set_side(Room::Direction::RIGHT, std::make_unique<Pass>(*m_rooms[i][j], *m_rooms[i][j+1]));
                m_rooms[i][j+1]->set_side(Room::Direction::LEFT, std::make_unique<Pass>(*m_rooms[i][j+1], *m_rooms[i][j]));
                int temp = row[j + 1];
                for (int k = 0; k < cols; k++) {
                    if (row[k] == temp) row[k] = row[j];
                }
            }
        }
        for (int j = 0; j < cols; j++) {
            if (rand() % 2) {
                int temp = 0;
                for (int k = 0; k < cols; k++) {
                    if (row[k] == row[j] && !m_rooms[i][k]->get_side(Room::Direction::DOWN)) temp++;
                }
                if (temp > 1) {
                    m_rooms[i][j]->set_side(Room::Direction::DOWN, std::make_unique<Wall>(*m_rooms[i][j]));
                    m_rooms[i+1][j]->set_side(Room::Direction::UP, std::make_unique<Wall>(*m_rooms[i+1][j]));
                }
                else {
                    m_rooms[i][j]->set_side(Room::Direction::DOWN, std::make_unique<Pass>(*m_rooms[i][j], *m_rooms[i+1][j]));
                    m_rooms[i+1][j]->set_side(Room::Direction::UP, std::make_unique<Pass>(*m_rooms[i+1][j], *m_rooms[i][j]));
                }
            }
            else {
                m_rooms[i][j]->set_side(Room::Direction::DOWN, std::make_unique<Pass>(*m_rooms[i][j], *m_rooms[i+1][j]));
                m_rooms[i+1][j]->set_side(Room::Direction::UP, std::make_unique<Pass>(*m_rooms[i+1][j], *m_rooms[i][j]));
            }
        }
        for (int j = 0; j < cols; j++) {
            if (m_rooms[i][j]->get_side(Room::Direction::DOWN)) row[j] = count++;
        }
    }
}
