#include "GameState.h"
#include "SelectState.h"
#include "../../../config.h"
#include "../Drawable/Entity/IEntity.h"
#include "ThemeManager.h"
#include "../Assets/Texture_Sound.h"

GameState::GameState(IStateManager &state_manager, const sf::VideoMode &video_mode, const std::string &window_title) :
        IState(state_manager), IWindowKeeper(config::GAME_VIDEO_MODE, window_title){
    if (!m_sound_ingame.openFromFile(ASSETS_PATH "ingame.mp3"))
        throw std::runtime_error("Failed to load sound ingame");
    if (!m_sound_win.openFromFile(ASSETS_PATH "win_game.mp3"))
        throw std::runtime_error("Failed to load sound win");
    if (!m_sound_lost.openFromFile(ASSETS_PATH "lost_game.mp3"))
        throw std::runtime_error("Failed to load sound lost");
    if (!m_screamer_music.openFromFile(ASSETS_PATH "screamer.mp3"))
        throw std::runtime_error("Failed to load screamer sound");
    m_sound_ingame.setLoop(true); ///> зацикливание музыки
    m_sound_win.setLoop(true);
    m_sound_lost.setLoop(true);
}

void GameState::event_handling() {
    sf::Event event{};
    GameContext& context = m_context_manager.get_current_context();
    bool mode_made = false;
    while (m_window.pollEvent(event)) {
        if (event.type == sf::Event::Closed)
            m_state_manager.set_next_state(std::make_unique<SelectState>(m_state_manager,  config::GAME_VIDEO_MODE, config::SELECT_LEVEL_TITLE));

        if (context.screamer_active) continue;

        if (event.type == sf::Event::KeyPressed && event.key.code == sf::Keyboard::Z)
            m_context_manager.restore_previous_context();

        if (event.type == sf::Event::KeyPressed && m_context_manager.get_current_context().state  == GameContext::INGAME) {
            if (!mode_made)
                mode_made = process_key_pressed(event.key.code);
        }
    }
}

bool GameState::process_key_pressed(sf::Keyboard::Key code) {
    GameContext& context = m_context_manager.get_current_context();

    if (code == sf::Keyboard::W || code == sf::Keyboard::A || code == sf::Keyboard::S || code == sf::Keyboard::D) {
        m_context_manager.save_current_context();
        switch (code) {
            case sf::Keyboard::W: context.pacman.move(Room::UP); break;
            case sf::Keyboard::S: context.pacman.move(Room::DOWN); break;
            case sf::Keyboard::A: context.pacman.move(Room::LEFT); break;
            case sf::Keyboard::D: context.pacman.move(Room::RIGHT); break;
        }
        return true;
    }

    if (code == sf::Keyboard::Up || code == sf::Keyboard::Down || code == sf::Keyboard::Left || code == sf::Keyboard::Right) {
        if (context.pacman2.get_location() != nullptr) {
            m_context_manager.save_current_context();
            switch (code) {
                case sf::Keyboard::Up: context.pacman2.move(Room::UP); break;
                case sf::Keyboard::Down: context.pacman2.move(Room::DOWN); break;
                case sf::Keyboard::Left: context.pacman2.move(Room::LEFT); break;
                case sf::Keyboard::Right: context.pacman2.move(Room::RIGHT); break;
            }
            return true;
        }
    }
    return false;
}
sf::Clock m_clock;

void GameState::update() {
    std::vector<std::unique_ptr<IGameEvent>> game_events;
    if (m_context_manager.get_current_context().state != GameContext::INGAME) return;
    GameContext& context = m_context_manager.get_current_context();

    if (context.screamer_active) {
        context.screamer_timer -= m_delta_time;
        if (context.screamer_timer <= 0.0f) context.screamer_active = false;
    }

    for (auto& ptr_entity : context.dynamic_objects) {
        ptr_entity->action();
        IVisitor* ptr_visitor = &(context.pacman);
        auto ptr_event = ptr_entity->accept(ptr_visitor);
        if (ptr_event)  game_events.emplace_back(std::move(ptr_event));

        if (context.pacman2.get_location() != nullptr) {
            IVisitor* ptr_visitor2 = &(context.pacman2);
            auto ptr_event2 = ptr_entity->accept(ptr_visitor2);
            if (ptr_event2)  game_events.emplace_back(std::move(ptr_event2));
        }
    }
    for (auto& static_ptr : context.static_objects) {
        IVisitor* ptr_visitor = &(context.pacman);
        auto ptr_event = static_ptr->accept(ptr_visitor);
        if (ptr_event)  game_events.emplace_back(std::move(ptr_event));

        if (context.pacman2.get_location() != nullptr) {
            IVisitor* ptr_visitor2 = &(context.pacman2);
            auto ptr_event2 = static_ptr->accept(ptr_visitor2);
            if (ptr_event2)  game_events.emplace_back(std::move(ptr_event2));
        }
    }

    if (context.static_objects.empty())
        game_events.emplace_back(std::move(std::make_unique<WinGame>()));

    for (const auto & game_event : game_events)
        game_event->handle(&m_context_manager.get_current_context());

    game_events.clear();
}

void GameState::render() {
    GameContext& context = m_context_manager.get_current_context();
    sf::Color backgroundColor;

    switch (context.state) {
        case GameContext::INGAME:
            if (m_current_music != Music::INGAME) {
                m_sound_lost.stop();
                m_sound_win.stop();
                m_sound_ingame.play();
                m_current_music = Music::INGAME;
            }
            backgroundColor = ThemeManager::Instance().getGameBackgroundInGameColor();
            break;
        case GameContext::LOST:
            if (m_current_music != Music::LOST) {
                m_sound_win.stop();
                m_sound_ingame.stop();
                m_sound_lost.play();
                m_current_music = Music::LOST;
            }
            backgroundColor = ThemeManager::Instance().getGameBackgroundLostColor();
            break;
        default:
            if (m_current_music != Music::WIN) {
                m_sound_lost.stop();
                m_sound_ingame.stop();
                m_sound_win.play();
                m_current_music = Music::WIN;
            }
            backgroundColor = ThemeManager::Instance().getGameBackgroundWinColor();
    }

    m_window.clear(backgroundColor);
    m_maze.draw_into(m_window);
    context.pacman.draw_into(m_window);
    if (context.pacman2.get_location() != nullptr)
        context.pacman2.draw_into(m_window);
    for (auto& obj :  context.static_objects)
        obj->draw_into(m_window);
    for (auto& el : context.dynamic_objects)
        el->draw_into(m_window);

    if (context.screamer_active) {
        if (m_screamer_music.getStatus() != sf::SoundSource::Playing) {
            m_sound_ingame.stop();
            m_sound_win.stop();
            m_sound_lost.stop();
            m_screamer_music.play();
        }

        sf::Sprite sprite;
        sprite.setTexture(MyAssets_one::GetScreamerTexture());

        sf::Vector2u windowSize = m_window.getSize();
        sf::Vector2u textureSize = sprite.getTexture()->getSize();

        sprite.setScale(
                static_cast<float>(windowSize.x) / textureSize.x,
                static_cast<float>(windowSize.y) / textureSize.y
        );

        const int shakeRange = 10;
        int offsetX = rand() % (shakeRange * 2 + 1) - shakeRange;
        int offsetY = rand() % (shakeRange * 2 + 1) - shakeRange;
        sprite.setPosition(static_cast<float>(offsetX), static_cast<float>(offsetY));

        m_window.draw(sprite);
    }
    else {
        if (m_screamer_music.getStatus() == sf::SoundSource::Playing)
            m_screamer_music.stop();
    }

    m_window.display();
}

bool GameState::do_step() {
    m_delta_time = m_clock.restart().asSeconds();
    event_handling();
    if (m_context_manager.get_current_context().state == GameContext::INGAME)
        update();
    render();
    return true;
}

void GameState::set_context(GameContext&& context) {
    m_context_manager.reset(std::move(context));
}

GameState::~GameState() {
    m_sound_lost.stop();
    m_sound_win.stop();
    m_sound_ingame.stop();
}
